/******************************************************************************
 * Author: Phil DeMonaco
 *
 * The primary source file for the RPSLS (Rock Paper Scissors Lizard Spock)
 * client. Curses support is required for operation. The user must specify
 * their desired name and the hostname of the server. Once two clients have
 * connected to a given server the match begins.  Ctrl-C to quit
 *
 * Networking code taken from Molay, "Understanding Unix/Linux Programming"
 * http://wps.prenhall.com/esm_molay_UNIXProg_1/
/******************************************************************************/

#include <curses.h>         // Provides nice termial functions
#include <errno.h>          // Needed to print connection errors
#include "errorHandler.h"   // Needed for standard error handling
#include "rpsls.h"          // set of defines for RPSLS
#include <signal.h>         // Signal handling functions
#include "socketlib.h"      // Socket connection library
#include "stddef.h"         // Generic defines (TRUE, FALSE, etc)
#include <stdio.h>          // fread, fprintf, etc...
#include <stdlib.h>         // exit, EXIT_SUCCESS, EXIT_FAILURE etc
#include <string.h>         // strlen, strcpy, strcmp etc

// Flags for configDisplay
#define SETUP 0
#define SHUTDOWN 1

struct point {  // simple struct containing x and y cordinates and a value
    int x;      // used to organize display data
    int y;
    int val;
};

typedef struct point POINT; // typedef for point allowing easier access

static int userChoice;              // the users choice in the last round
static POINT curs;                  // used to store the cursors current pos
static POINT userScore;             // display location for the users score
static POINT opntScore;             // display location for the opponents score
static POINT prompt;                // display location for the prompt
static POINT round;                 // display location for the current round
static char hostname[HOSTLEN];      // name of the server
static char opponent[MAX_ARG_SIZE]; // opponents name
static FILE *server;                // formatted file descriptor for the server
static int socketFD;                // file descriptor for the server connection
static char username[MAX_ARG_SIZE]; // name of the user
static WINDOW *userWindow;          // pointer to the current curses window

static char choices [5][9] = // array of possible user choices
        { "rock", "paper", "scissors", "lizard", "spock" };

static char outcomes[5][5][12] = // matrix of user choice outcomes
       {{      "Tie",   "covered",     "crushed",     "crushed", "vaporized"},
        {  "covered",       "Tie",         "cut",       "eaten", "disproved"},
        {  "crushed",       "cut",         "Tie", "decapitated",   "smashed"},
        {  "crushed",     "eaten", "decapitated",         "Tie",  "poisoned"},
        {"vaporized", "disproved",     "smashed",    "poisoned",       "Tie"}};

// function stubs
void play();
void configDisplay(int operation);
void endgame(const char* cmd);
void handler(int signal);
void packServerRead(char *cmd, char *arg);
void printErrorC(char *message, void *param, int paramType, int fatal);
void printInterface();
void promptHost();
void promptName();
char *readUserInput(char *dest, int maxSize);
void displayRound(char *cmd, char *arg);
void sendMSG(const char *arg);
void setOpponent(char *arg);
void turn();

/***********************************************************************
 * Function: main
 * Purpose: Handle connection to the server and initiate game code. Two
 *              optional parameters may be provided: hostname and username.
 *              All further arguments are ignored.
 *
 * Parameter(s):
 *  (int)       argc:       number of arguments
 *  (char**)    argv:       array of command line arguments
 *
 * Returns: int         Was the run successful?
/***********************************************************************/
int main(int argc, char** argv)
{
    char choice = NULLCHR; // prompt buffer if the server connection fails

    signal(SIGINT, handler); // configure SIGINT such that we can exit
    configDisplay(SETUP); // setup curses

    if (argc > 1) strncpy(hostname, argv[1], HOSTLEN); // if avaliable use cmd
    else promptHost(); // line data for the server connection otherwise prompt

    if (argc > 2) strncpy(username, argv[2], MAX_ARG_SIZE); // if avaliable use
    else promptName(); // cmd line data for the user name otherwise prompt

    // attempt to connect to the server
    while ((socketFD = connectToServer(hostname, DEFAULT_PORT)) == ERROR)
    {
        clear(); // if something went wrong clear the screen
        mvprintw(0,0,"Failed to connect to '%s': %s\n", // tell the user about
                hostname, strerror(errno));

        do // wait for a valid user choice
        {
            clrtoeol(); // clear the line
            getyx(userWindow, curs.y, curs.x); // get the cursor position
            mvprintw(curs.y, 0, "Try different server name (y/n): "); // prompt
            choice = getch(); // retrieve the user's choice
        } while (choice != YES && choice != NO);

        // carry out the appropriate action based on the user's choice
        switch(choice)
        {
            case YES: // the user wants to try a different server
                move(curs.y+1,0); // go to the next line
                promptHost(); // prompt for a new server name
                choice = FALSE; // clear choice
                break;
            case NO: // quit
                configDisplay(SHUTDOWN);
                exit(EXIT_SUCCESS);
        }
    }



    play(); // start the game

    configDisplay(SHUTDOWN); // shutdown curses & restore the terminal
    return (EXIT_SUCCESS); // tell the os everything went well
}

/***********************************************************************
 * Function: configDisplay
 * Purpose: configures the display elements of the terminal. In particular
 *          curses is either setup or shutdown depending on the input. This
 *          method disables echo, clears the screen, and changes to non
 *          canonical mode. If SHUTDOWN is specified curses is disabled
 *          restoring all previous terminal settings.
 *
 * Parameter(s):
 *  (int)    operation:   what task should be performed? SETUP | SHUTDOWN
/***********************************************************************/
void configDisplay(int operation)
{
    if (operation == SETUP)
    {
        userWindow = initscr(); // initialize curses
        clear();   // clear the display
        echo();    // its nice to see what we chose
        refresh(); // refresh the display (apply clear)
    }
    else
        endwin();  // revert to previous terminal settings & exit curses
}

/***********************************************************************
 * Function: displayRound
 * Purpose: prints the outcome of the last round as relayed by the server
 *              in one of two commands, ROUND_LOSS or ROUND_WIN
 *
 * Parameter(s):
 *  (char*) cmd:   the round command either ROUND_LOSS or ROUND_WIN
 *  (char*) arg:   choice of the other user
/***********************************************************************/
void displayRound(char *cmd, char *arg)
{
    int rwon;       // boolean value to store the round outcome
    int opntChoice; // the opponents choice

    opntChoice = atoi(arg); // read the opponents choice
    rwon = strcmp(cmd, ROUND_WIN) == EQUAL; // check the command for round status

    if (rwon) userScore.val++; // if we won increment our score
    else if (opntChoice != userChoice)
        opntScore.val++; // if we lost but didn't tie increment the opponent's

    mvprintw(userScore.y, userScore.x, "%d", userScore.val); // print our score
    mvprintw(opntScore.y, opntScore.x, "%d", opntScore.val); // print the opnt's

    move(prompt.y,prompt.x); // move to the prompt line
    clrtoeol(); // clear it

    // if the round was not a tie print the outcome
    if (opntChoice != userChoice)
        printw(rwon ? WIN_STR : // If we won print the win string else
            userChoice != 2 ? // determine the correct loss string to print
                LOSS_STR : LOSS_STR_SCISSORS,
            choices[userChoice], outcomes[userChoice][opntChoice],
            opponent, choices[opntChoice]);
    else addstr("The round was a tie"); // if not a tie tell the user about it

    move(prompt.y+1,0); // move to the next line
    addstr("Press any key to continue..."); // prompt the user to continue
    refresh(); // refresh the display
    getch(); // wait for any key press
    move(prompt.y+1,0); // move to the beginning of the prompt line
    clrtoeol(); // clear it
}

/***********************************************************************
 * Function: endgame
 * Purpose: End the game under one of four conditions: WIN, LOSS, FORFEIT,
 *              and LOST_SERVER.  Tell the user about these conditions and
 *              close gracefully
 *
 * Parameter(s):
 *  (char*) cmd:   the endgame command
/***********************************************************************/
void endgame(const char* cmd)
{    
    move(prompt.y,prompt.x); // move to the prompt line
    clrtobot(); // clear the display below the current line

    // print the end game condition
    if (strcmp(cmd, WIN) == EQUAL) // we won
        addstr("You are victorious!");
    else if (strcmp(cmd, LOSS) == EQUAL) // we lost
        addstr("You have been defeated!");
    else if (strcmp(cmd, FORFEIT) == EQUAL) // the other player quit
        printw("%s has forfeit the match! You are victorious!",
                strlen(opponent) > 1 ? opponent : "Your opponent");
    else if (strcmp(cmd, LOST_SERVER) == EQUAL) // we lost connection to the
        addstr("Connection to server lost!"); // server

    move(prompt.y+1,0); // move down one line
    addstr("Press any key to exit..."); // prompt the uesr and wait
    refresh();
    getch(); // wait for any input

    configDisplay(SHUTDOWN); // shutdown curses
    fclose(server);          // attempt to close the FILE
    close(socketFD);         // close the connection socket
    exit(EXIT_SUCCESS);      // exit the program
}

/***********************************************************************
 * Function: handler
 * Purpose: Handle SIGINT by gracefully shutting down curses, closing the
 *              server socket, and exiting
 *
 * Parameter(s):
 *  (int) signal:   signal which triggered this interrupt, unused
/***********************************************************************/
void handler(int signal)
{
    configDisplay(SHUTDOWN); // shutdown curses
    fclose(server);          // attempt to close the FILE
    close(socketFD);         // close the socket
    exit(EXIT_SUCCESS);      // exit the program
}

/***********************************************************************
 * Function: packServerRead
 * Purpose: Pack a message from the server into the given buffers
 *
 * Parameter(s):
 *  (char*) cmd:   the round command either ROUND_LOSS or ROUND_WIN
 *  (char*) arg:   choice of the other user
/***********************************************************************/
void packServerRead(char *cmd, char *arg)
{
    int argLen = 0; // length of the argument portion
    char cmdBuf[HEAD_SIZE]; // buffer for the command

    memset(cmd, NULLCHR, CMD_SIZE);     // zero the three buffers
    memset(cmdBuf, NULLCHR, HEAD_SIZE);
    memset(arg, NULLCHR, MAX_ARG_SIZE);

    fread(cmdBuf, sizeof(char), HEAD_SIZE-1, server); // read the command

    memcpy(cmd, cmdBuf, CMD_SIZE-1); // copy the command into cmdBuf
    argLen = atoi(cmdBuf+CMD_SIZE-1); // convert arglen to an int
    
    fread(arg, sizeof(char), argLen, server); // read the argument

    if (ferror(server)) // if the connection errored tell someone about it
        printErrorC("RPSLS client: A server read error occured: ",
                NULLCHR, STRING, TRUE);

    if (feof(server)) endgame(LOST_SERVER); // if eof was reached run endgame

    move(10,0);
    clrtoeol();
    printw("Read '%s%s'", cmdBuf, arg);
    refresh();
}

/***********************************************************************
 * Function: play
 * Purpose: Waits for server commands and performs the specified actions.
/***********************************************************************/
void play()
{
    char cmd[CMD_SIZE];     // buffer for server command codes
    char arg[MAX_ARG_SIZE]; // argument buffer

    printInterface(); // setup the game interface

    // attempt to open the server file descriptor as a FILE
    if ((server = fdopen(socketFD, READ_WRITE)) == NULL)
        printErrorC("RPSLS client: failed to open file descriptor '%d': ",
                &socketFD, INT, TRUE);

    while(TRUE) // play for as long as it takes
    {
        packServerRead(cmd, arg); // read a command from the server

        // perform the action specified by the server command
        if ((strcmp(cmd, GET_NAME)) == EQUAL)
            sendMSG(username); // send username to the server
        else if ((strcmp(cmd, LOSS) == EQUAL) || (strcmp(cmd, WIN) == EQUAL) ||
                 (strcmp(cmd, FORFEIT) == EQUAL))
            endgame(cmd); // end the game with the given condition
        else if ((strcmp(cmd, ROUND_WIN) == EQUAL) ||
                 (strcmp(cmd, ROUND_LOSS) == EQUAL))
            displayRound(cmd, arg); // display the outcome of the round
        else if (strcmp(cmd, SET_OPPONENT) == EQUAL)
            setOpponent(arg); // store and display the opponents name
        else if (strcmp(cmd, TURN) == EQUAL)
            turn(arg); // initate a new turn
    }
}

/***********************************************************************
 * Function: printErrorC
 * Purpose: Gracefully closes curses before calling printError with
 *              the given parameters
 *
 * Parameter(s):
 *  (char*) message:    the message to be printed
 *  (void*) param:      an optional parameter for the given message
 *  (int)   paramType:  parameter type; STRING, CHAR, or INT
 *  (int)   fatal:      boolean flag indicating if the program should exit
/***********************************************************************/
void printErrorC(char *message, void *param, int paramType, int fatal)
{
    configDisplay(SHUTDOWN);
    printError(message, param, paramType, fatal);
}

/***********************************************************************
 * Function: printInterface
 * Purpose: Prints several basic interface components
/***********************************************************************/
void printInterface()
{
    clear(); // clear the screen
    addstr("RPSLS v1.0"); // print the game name
    mvprintw(1,0,"Host: '%s'", hostname); // print the server's name
    mvaddstr(2,0,"Waiting for opponent..."); // we're waiting for another player
    refresh(); // flush this data to the display
}

/***********************************************************************
 * Function: promptHost
 * Purpose: Prompt's the user for a server name. No longer that HOSTLEN
/***********************************************************************/
void promptHost()
{
    printw("Server (Max %d): ", HOSTLEN); // print the prompt
    refresh(); // flush it to the display
    readUserInput(hostname, HOSTLEN); // read the input
}

/***********************************************************************
 * Function: promptName
 * Purpose: Prompt's the user for their username. No longer that MAX_ARG_SIZE
/***********************************************************************/
void promptName()
{
    printw("Client (Max %d): ", MAX_ARG_SIZE); // print the prompt
    refresh();  // flush it to the display
    readUserInput(username, MAX_ARG_SIZE); // read the input
}

/***********************************************************************
 * Function: readUserInput
 * Purpose: reads user input of the given length into the specified buffer
 *
 * Parameter(s):
 *  (char*) dest:       buffer in which to store the data
 *  (int)   maxSize:    maximum size in bytes of the read data including \0
/***********************************************************************/
char *readUserInput(char *dest, int maxSize)
{
    if(wgetnstr(userWindow, dest, maxSize-1) == ERR) // read the input
        printErrorC("RPSLS client: a read error occured: ",
                NULLCHR, STRING, TRUE); // if something goes wrong tell user
    return dest; // return the buffer
}

/***********************************************************************
 * Function: sendMSG
 * Purpose: sends a message to the server, messages must be shorter than
 *              MAX_ARG_SIZE-1 (the full size is including termination)
 *
 * Parameter(s):
 *  (const char*) dest:       message to be sent
/***********************************************************************/
void sendMSG(const char *arg)
{
    int argLen; // length of the message
    char msg[MAX_ARG_SIZE]; // message buffer

    memset(msg, NULLCHR, MAX_ARG_SIZE); // zero the buffer

    strcpy(msg, arg); // copy the message into the buffer
    argLen = strlen(msg); // figure out how long it is
    argLen = argLen < MAX_ARG_SIZE-1 // if its too long truncate it
            ? argLen : MAX_ARG_SIZE-1;
    // send the message to the server the beggining of the message is the length
    fprintf(server, "%02d%s", argLen, msg);
    fflush(server);

    move(10,0);
    clrtoeol();
    printw("Sent '%02d%s'", argLen, msg);
    refresh();
}

/***********************************************************************
 * Function: setOpponent
 * Purpose: Stores and displays the given value as the opponents name
 *
 * Parameter(s):
 *  (char*) arg:       opponent name
/***********************************************************************/
void setOpponent(char *arg)
{
    memset(opponent, NULLCHR, MAX_ARG_SIZE); // initalize opponent to zeros
    strncpy(opponent, arg, MAX_ARG_SIZE-1); // store the name

    move(2,0); // move to the third line
    clrtoeol(); // clear it
    printw("Score %s: ", username); // Print the score header followed by user
    getyx(userWindow, userScore.y, userScore.x); // capture the position
    printw("0 %s: ", opponent); // print a zero and the opponent name
    getyx(userWindow, opntScore.y, opntScore.x); // capture the opponent pos
    addstr("0 Round: "); // print a zero and round
    getyx(userWindow, round.y, round.x); // capture the round position
    addstr("0\n-------------------------------"); // print a zero and a divider
    printw("\nRock '%c' Paper '%c' Scissors '%c'\nLizard '%c' Spock '%c'\nCtrl-C"
            " to quit", ROCK, PAPER, SCISSORS, LIZARD, SPOCK); // print choices
    addstr("\n-------------------------------\n"); // print another dividing line
    getyx(userWindow, prompt.y, prompt.x); // capture the prompt location
    refresh(); // flush to the display
}

/***********************************************************************
 * Function: turn
 * Purpose: Executes a turn. This includes printing the current round num
 *              prompting the user for a choice and sending that choice
 *              to the server
 *
 * Parameter(s):
 *  (char*) arg:       string containing the round number
/***********************************************************************/
void turn(char *arg)
{
    char choice[2];  // string to store the user's choice
    int cont = TRUE; // flag determining if good input was recieved

    round.val = atoi(arg); // parse the round number
    memset(choice, NULLCHR, 2); // zero choice

    mvprintw(round.y, round.x, "%d", round.val); // print the round number
    move(prompt.y,prompt.x); // move to the prompt line
    clrtoeol(); // clear it
    addstr("Choice: "); // print the prompt
    getyx(userWindow, curs.y, curs.x); // get the current position

    while (cont) // while the input is bad
    {
        move(curs.y, curs.x); // move to the prompt point
        clrtoeol(); // clear the rest of the lint
        refresh(); // flush to display

        switch(choice[0] = getch()) // retrieve user input
        {
            case ROCK:          // if its one of the five valid choices
            case PAPER:
            case SCISSORS:
            case LIZARD:
            case SPOCK:
                cont = FALSE; // set cont to false and break
                break;
            default: // otherwise keep going
                break;
        }
    }

    sendMSG(choice); // send the choice as a message
    userChoice = atoi(choice); // parse and store the choice
}
