/******************************************************************************
 * Author: Phil DeMonaco
 *
 * The primary source file for the RPSLS (Rock Paper Scissors Lizard Spock)
 * server. This server listens for client connections on socket 25000 by default.
 * Once two clients have connected the server creates a child process to execute
 * the game logic.
 *
 * Networking code taken from Molay, "Understanding Unix/Linux Programming"
 * http://wps.prenhall.com/esm_molay_UNIXProg_1/
/******************************************************************************/

#include <errno.h>          // check if a system error needs to be printed
#include "errorHandler.h"   // provides basic error handling functionality
#include "rpsls.h"          // set of defines for RPSLS
#include <sys/select.h>     // provides access to select...
#include <signal.h>         // signal, SIGCHLD & SIGINT
#include "socketlib.h"      // basic stream socket library
#include "stddef.h"         // set of standard defines (TRUE, FALSE, etc...)
#include <stdio.h>          // fprintf, fread, FILE, etc...
#include <stdlib.h>         // exit, EXIT_SUCCESS, EXIT_FAILURE, etc...
#include <string.h>         // standard string operations, strlen, strcpy, etc..
#include <sys/wait.h>       // needed for waitpid

// Parent Variables
static int socket; // file descriptor for the listening socket

// Child Variables
static FILE *clients[2];    // set of file streams for the two clients

// function stubs
void endGame();
int evalRound(char *choice);
void forfiet();
FILE *getClient();
void handler(int signal);
void packChoice(char *choice, int roundNum);
void play();
char *readClient(FILE *stream, char *dest);
void sendCMD(FILE *sendCMD, const char *cmd, const char *arg);
void sendResult(int result, char *choice);
void transferNames();
void win(int result);

/***********************************************************************
 * Function: main
 * Purpose: Configures a listening socket on port 25000 accepts client
 *              connections & forks game handling processes
 *
 * Parameter(s):
 *  (int)       argc:       number of arguments (unused)
 *  (char**)    argv:       array of command line arguments (unused)
 *
 * Returns: int         Was the run successful?
/***********************************************************************/
int main(int argc, char** argv)
{
    int PID;    // stores the pid when the program forks

    if ((socket = createServerSocket(DEFAULT_PORT)) == ERROR) // attempt to
        printError("RPSLS server: failed to create socket: ", // create the
                NULLCHR, STRING, TRUE);     // socket

    signal(SIGCHLD, handler); // configure the signal handler for SIGCHLD
    signal(SIGINT, handler); // and SIGINT

    // main loop waits for two clients then starts the game
    while(TRUE)
    {   
        clients[CLIENT1] = getClient(); // wait for the first client
        if (clients[CLIENT1] == NULL && errno == EINTR) 
            continue; // if accept was interrupted abort

        clients[CLIENT2] = getClient(); // wait for the second client
        if (clients[CLIENT2] == NULL && errno == EINTR) 
            continue; // if accept was interrupted abort

        // if the clients don't exist error
        if (clients[CLIENT1] == NULL || clients[CLIENT2] == NULL)
            printError("RPSLS server: failed to connect to clients: ",
                    NULLCHR, STRING, TRUE);

        if ((PID = fork()) == ERROR) // attempt to fork
            printError("RPSLS server: failed to fork: ", NULLCHR, STRING, TRUE);
        
        if (PID == 0) play(); // if we are the fork begin the game

        memset(clients, NULLCHR, MAX_CLIENTS); // zero the clients
    }

    close(socket); // we're done, close the connection
    
    return (EXIT_SUCCESS); // tell the os everything went well
}

/***********************************************************************
 * Function: endGame
 * Purpose: Closes the two client connections and exits. only called by
 *              children.
/***********************************************************************/
void endGame()
{
    fclose(clients[CLIENT1]); // close the two client streams
    fclose(clients[CLIENT2]);
    exit(EXIT_SUCCESS); // exit gracefully
}

/***********************************************************************
 * Function: evalRound
 * Purpose: calculates the winner of the current round based on the provided
 *              client choices.
 *
 * Parameter(s):
 *  (char*) choice:     array containing the choices of users 1 and 2
 * Returns: int:    the result (CLIENT1, CLIENT2 or TIE)
/***********************************************************************/
int evalRound(char *choice)
{
    if (choice[CLIENT1] == choice[CLIENT2]) 
        return TIE; // if the choices are equal tie

    switch (choice[CLIENT1]) // return the outcome based on clients 1 and 2
    {
        case ROCK: // ROCK beats LIZARD & SCISSORS
            return choice[CLIENT2] == LIZARD || choice[CLIENT2] == SCISSORS
                    ? CLIENT1 : CLIENT2;
        case PAPER: // PAPER beats ROCK and SPOCK
            return choice[CLIENT2] == ROCK || choice[CLIENT2] == SPOCK
                    ? CLIENT1 : CLIENT2;
        case SCISSORS: // SCISSORS beats PAPER and LIZARD
            return choice[CLIENT2] == PAPER || choice[CLIENT2] == LIZARD
                    ? CLIENT1 : CLIENT2;
        case LIZARD: // LIZARD beats PAPER and SPOCK
            return choice[CLIENT2] == PAPER || choice[CLIENT2] == SPOCK
                    ? CLIENT1 : CLIENT2;
        case SPOCK: // SPOCK beats ROCK and SCISSORS
            return choice[CLIENT2] == ROCK || choice[CLIENT2] == SCISSORS
                    ? CLIENT1 : CLIENT2;
    }
}

/***********************************************************************
 * Function: forfiet
 * Purpose: one of the users disconnected, tell the other user and exit
/***********************************************************************/
void forfiet()
{
    int client; // index for clients

    for (client = 0; client < MAX_CLIENTS; client++) // we dont know which
        sendCMD(clients[client], FORFEIT, NULL); // so tell everyone
    
    fprintf(stderr, "RPSLS server: A client disconnected\n"); // print an error
    endGame(); // exit the game
}

/***********************************************************************
 * Function: getClient
 * Purpose: retrieve a new client stream from the socket
 * Returns: FILE*: a file stream pointing to the new client
/***********************************************************************/
FILE *getClient()
{
    int clientFD; // file descriptor for the client
    FILE *client; // file stream for the client
    
    if ((clientFD = accept(socket, NULL, NULL)) == ERROR) // attempt to accept
        return NULL; // a connection, if something goes wrong return NULL
    if ((client = fdopen(clientFD, READ_WRITE)) == NULL) // attempt to open
        return NULL; // the file descriptor as a file stream
    
    return client; // return the new file stream
}

/***********************************************************************
 * Function: handler
 * Purpose: basic singal handler to deal with SIGCHLD and SIGINT.
 *
 * Parameter(s):
 *  (int)   signal: the signal which triggered the call
/***********************************************************************/
void handler(int signal)
{
    switch(signal) // switch to the action based on the given signal
    {
        case SIGCHLD: // SIGCHLD allows us to clean up zombies
            while (waitpid(-1, NULL, WNOHANG) > 0); // clean any that exist
            break;
        case SIGINT: // exit if SIGINT is given
            close(socket); // close the socket
            exit(EXIT_SUCCESS); // exit
    }
}

/***********************************************************************
 * Function: packChoice
 * Purpose: retrieve the clients choices for this round & pack them in
 *              the provided character array
 *
 * Parameter(s):
 *  (char*) choice:     array in which the choices are to be stored
 *  (int)   roundNum:   number of this round (to tell the clients)
/***********************************************************************/
void packChoice(char *choice, int roundNum)
{
    int client; // index into clients
    char arg[MAX_ARG_SIZE]; // buffer in which to store the message arguments

    memset(arg, NULLCHR, MAX_ARG_SIZE); // zero arg prior to use
    sprintf(arg, "%d", roundNum); // insert the round number into arg

    for (client = 0; client < MAX_CLIENTS; client++) // tell each client to
        sendCMD(clients[client], TURN, arg);         // execute a turn

    for (client = 0; client < MAX_CLIENTS; client++) // read and store the choice
    {                                                // from each client
        readClient(clients[client], arg);
        choice[client] = arg[0];
    }
}

/***********************************************************************
 * Function: transferNames
 * Purpose: perform the inital transfer of names between the clients
/***********************************************************************/
void transferNames()
{
    int client; // index into clients
    char name[MAX_CLIENTS][MAX_ARG_SIZE]; // buffer for the client names

    for (client = 0; client < MAX_CLIENTS; client++) // request each client's
        sendCMD(clients[client], GET_NAME, NULL);   // user name

    for (client = 0; client < MAX_CLIENTS; client++) // read the name of each
        readClient(clients[client], name[client]);  // client

    sendCMD(clients[CLIENT1], SET_OPPONENT, name[CLIENT2]); // send the clients
    sendCMD(clients[CLIENT2], SET_OPPONENT, name[CLIENT1]); // name to its opnt
}

/***********************************************************************
 * Function: play
 * Purpose: primary game logic method for the server. controls program flow
/***********************************************************************/
void play()
{
    int result = TIE;   // the result of the current round
    int roundNum = 0;   // the current round number
    char roundChoice[MAX_CLIENTS];  // the player's choices for this round
    int score[MAX_CLIENTS]; // the current score of each player

    score[CLIENT1] = score[CLIENT2] = 0; // zero the player scores

    transferNames(clients); // perform the inital name transfer

    while(TRUE)
    {
        roundNum++; // increment the round number
        packChoice(roundChoice, roundNum); // retrieve the choices
        result = evalRound(roundChoice); // determine the round winner

        if (result != TIE) score[result]++; // if it wasn't a tie increment score

        sendResult(result, roundChoice);    // send out the result

        if (score[CLIENT1] > 1) win(CLIENT1); // check to see if either player
        else if (score[CLIENT2] > 1) win(CLIENT2); // won the game
    }
}

/***********************************************************************
 * Function: readClient
 * Purpose: read a message from the given file stream into the provided
 *              buffer. Messages can be no longer than MAX_ARG_LEN
/***********************************************************************/
char *readClient(FILE *stream, char *dest)
{
    int argLen;     // length of the message
    char lenBuf[3]; // buffer to hold the length string
    
    memset(lenBuf, NULLCHR, LEN_SIZE); // zero both buffers
    memset(dest, NULLCHR, MAX_ARG_SIZE);

    fread(lenBuf, sizeof(char), LEN_SIZE-1, stream); // read in the length header
    argLen = atoi(lenBuf); // parse it to an integer so we can read it

    // using the argLen read the message into dest
    fread(dest, sizeof(char), argLen, stream);

    if (feof(stream)) forfiet();    // if EOF was reached someone forfeit
    if (ferror(stream))             // if an error occured tell someone
        printError("RPSLS server: failed to read from client!",
                NULLCHR, STRING, TRUE);

    if (strcmp(dest, QUIT) == EQUAL) forfiet(); //if a client is quitting forfeit

    return dest; // return the buffer
}

/***********************************************************************
 * Function: sendCMD
 * Purpose: send a command to the given stream
 *
 * Parameter(s):
 *  (FILE*)         stream:     file stream to send the command
 *  (const char*)   cmd:        the command to be send (CMD_SIZE or shorter)
 *  (const char*)   arg:        argument to be sent (MAX_ARG_SIZE or shorter)
/***********************************************************************/
void sendCMD(FILE *stream, const char *cmd, const char *arg)
{
    char argBuf[MAX_ARG_SIZE];          // buffer for the argument string
    int argLen = 0;                     // length of the arg
    int streamFD = fileno(stream);      // file descriptor for the stream
    static fd_set set;                  // file descriptor set
    static struct timeval readTimeout;  // timeval for select

    readTimeout.tv_sec = 0; // we dont want to wait for seconds
    readTimeout.tv_usec = 1000; // 1 millisecond is long enough

    FD_ZERO(&set);  // zero the set
    FD_SET(streamFD, &set); // add streamFD to the set
    if ((select(streamFD+1, &set, NULL, NULL, &readTimeout)) > 0 // wait for
            && strcmp(cmd, FORFEIT) != EQUAL)   // readTimeout on the stream
        readClient(stream, argBuf); // if theres something to be read, read it
    
    if (arg) strcpy(argBuf, arg); // if arg isn't NULL copy it into argBuf
    else memset(argBuf, NULLCHR, MAX_ARG_SIZE); // otherwise null it
    argLen = strlen(argBuf); // determine the length of the argument
    argLen = argLen < MAX_ARG_SIZE-1 ? // make sure it doesn't go beyond
        argLen : MAX_ARG_SIZE-1; // the maximum argument length

    fprintf(stream, "%s%02d%s", cmd, argLen, argBuf); // write our data
    fflush(stream); // make sure the data is written now

    if (ferror(stream)) // if something goes wrong tell someone
        printError("RPSLS server: failed to write to client: ",
                NULLCHR, STRING, TRUE);
}

/***********************************************************************
 * Function: sendResult
 * Purpose: send the result of the current round to the clients. The command
 *              indicates win or loss while the argument is the other user's
 *              choice
 *
 * Parameter(s):
 *  (int)    result:    result of the current round
 *  (char*)  choice:    array of user choices
/***********************************************************************/
void sendResult(int result, char *choice)
{
    int client; // index into clients
    char arg[MAX_ARG_SIZE]; // buffer for the result argument

    for (client = 0; client < MAX_CLIENTS; client++) // populate arg with
    {                                                // the opponents choice
        sprintf(arg, "%c", choice[client == CLIENT1 ? CLIENT2 : CLIENT1]);
        sendCMD(clients[client], (client == result ? ROUND_WIN : ROUND_LOSS),
                arg); // then send the appropriate message
    }
}

/***********************************************************************
 * Function: win
 * Purpose: indicate victory or defeat to each client based on the outcome
 *              of the current round then exit.
 *
 * Parameter(s):
 *  (int)    result:    result of the current round (who won)
/***********************************************************************/
void win(int result)
{
    int client;         //index into clients

    // indicate victory or defeat to each client
    for (client = 0; client < MAX_CLIENTS; client++)
        sendCMD(clients[client], client == result ? WIN : LOSS, NULL);
    
    endGame(); // exit the game gracefully
}
