/******************************************************************************
 * Author: Phil DeMonaco
 *
 * A socket library based heavily on the material provided in chapter 11.  This
 * allows connections between clients & servers to be configured with relative
 * ease.
 *
 * Networking code taken from Molay, "Understanding Unix/Linux Programming"
 * http://wps.prenhall.com/esm_molay_UNIXProg_1/ 
/******************************************************************************/

#include <netdb.h>      // gethostbyname, hostent, etc...
#include <netinet/in.h> // sockaddr_and others...
#include <sys/types.h>  // socket, bind, etc..
#include <sys/socket.h> // socket, listen, bind
#include "socketlib.h"  // header for this libarary
#include "stddef.h"     // generic defines (TRUE, FALSE, etc...)
#include <string.h>     // memset, memcpy, etc
#include <unistd.h>     // gethostname

#define BACKLOG 1       // indicates the maximum number of queued clients

/***********************************************************************
 * Function: connectToServer
 * Purpose: basic function which creates a connection to the specified
 *              host at the specified port
 *
 * Parameter(s):
 *  (char*) host:   name of the desired host
 *  (int)   portNum:    port number at which to connect
 *
 * Returns: int         file descriptor for the new connection
/***********************************************************************/
int connectToServer(char *host, int portNum)
{
    int socketFD;   // file descriptor for the connection
    struct sockaddr_in servAddr;    // struct to store the server inet address
    struct hostent *servHostEnt;    // server host entry (Generated by DNS)

    // attempt to create an INETv4 stream socket. if something goes wrong error
    if ((socketFD = socket(AF_INET, SOCK_STREAM, 0)) == ERROR) return ERROR;
    
    memset(&servAddr, NULLCHR, sizeof(servAddr)); // zero servAddress
    // retrieve the entry for host. tell the user if it cant be found
    if ((servHostEnt = gethostbyname(host)) == NULL) return ERROR;

    // copy the entrys address to the socket
    memcpy((struct sockaddr *)&servAddr.sin_addr, servHostEnt->h_addr_list[0],
            servHostEnt->h_length);
    servAddr.sin_port = htons(portNum);   // store port address in network form
    servAddr.sin_family = AF_INET;        // specify socket family (INETv4)

    if (connect(socketFD, (struct sockaddr *)&servAddr, // attempt to connect
            sizeof(servAddr)) != 0) return ERROR;       // to the server

    return socketFD;    // return the new socket
}

/***********************************************************************
 * Function: createServerSocket
 * Purpose: create a new server socket listening at the given port number
 *
 * Parameter(s):
 *  (int)   portNum:    port number at which to listen
 *
 * Returns: int         file descriptor for the new connection
/***********************************************************************/
int createServerSocket(int portNum)
{
    return createServerSocket_p(portNum, BACKLOG);  // call the primary method
}

/***********************************************************************
 * Function: createServerSocket_p
 * Purpose: more powerful socket creation fucntion with provisions to
 *              specify the number of clients which can queue while
 *              connections are being established
 *
 * Parameter(s):
 *  (int)   portNum:    port number at which to listen
 *  (int)   backlog:    number of clients allowed to queue
 *
 * Returns: int         file descriptor for the new connection
/***********************************************************************/
int createServerSocket_p(int portNum, int backlog)
{
    struct hostent *hostEntry;      // host entry for this machine
    char hostName[HOSTLEN];         // name of the host
    struct sockaddr_in socketAddr;  // inet address of the socket
    int socketID;                   // file descriptor for the socket

    // attempt to create an INETv4 stream socket. if something goes wrong error
    if ((socketID = socket(PF_INET, SOCK_STREAM, 0)) == ERROR) return ERROR;

    memset(&socketAddr, NULLCHR, sizeof(socketAddr)); // zero socketAddr
    gethostname(hostName, HOSTLEN);        // retrieve the servers hostname
    hostEntry = gethostbyname(hostName);   // use hostName to retrieve the entry

    memcpy((void*)&socketAddr.sin_addr, // copy the host address to the
            (void*)hostEntry->h_addr_list[0], hostEntry->h_length); // socket
    socketAddr.sin_port = htons(portNum);   // store port address in network form
    socketAddr.sin_family = AF_INET;        // specify socket family (INETv4)

    if (bind(socketID, (struct sockaddr *)&socketAddr, // attempt to bind the
            sizeof(socketAddr)) != 0) return ERROR; // socket data to socketID

    if (listen(socketID, backlog) != 0) // begin listening for connections
        return ERROR;

    return socketID; // return the new file descriptor 
}
